#if defined WIN32
#pragma warning(disable : 4786)
#endif

#include "Module.h"

#include "Create.h"
#include "Command_template.h"
#include "Model.h"
#include "Project.h"
#include "Model_kit.h"
#include "Point_3D_impl.h"
#include "Lock.h"

#include "Force_field.h"
//#include "CC_energy.h"//fix
//#include "Model_energy_as_differentiable_and_retrievable_coordinates.h"
#include "Model_adapter.h"
#include "Min.h"

#include "Timer.h"
#include "User.h"
#include "Log.h"
#include "Peep.h"

#include "DbC.h"

#include "Panel_optimize.h"

#include <math.h>

namespace MM
{
class Module_min : public Module
{
    static Module_min       singleton_;
                            Module_min();
    void                    init ();

public:
    void                    optimize ();
    void                    clashes  ();
};
Module_min  Module_min::singleton_;

Module_min::Module_min() : Module ("Compute") { }

void Module_min::
init ()
{
    //==========
    // Min
    //==========

    adopt_command (create_command (*this, &Module_min::optimize))
        .set_menu        ("&Compute", "&Optimize")
        .set_accelerator ("Ctrl+P")
        .set_tool_tip    ("Optimize")
        .set_icon        ("Compute",  "optimize.png");//*/

    //insert_menu_separator ("&Compute");

    adopt_command (create_command (*this, &Module_min::clashes))
        .set_menu        ("&Build", "&Remove clashes")
        .set_tool_tip    ("Short optimization to remove atoms clashes (20 iterations)")
        .set_icon        ("Build",  "clash.png");//*/

}

void Module_min::
clashes ()
{
    struct Process_events : public Command
    {
        int i_;

        Process_events () : i_(0) {}

        void execute () 
        {
            if (++i_ % 1 == 0)
                qApp->processEvents(QEventLoop::ExcludeSocketNotifiers);
        }
    };

    Timer timer;
    Text  time;

    Model & model = Project::singleton().current_model();
    Model_adapter adapter (model);

    try
    {
        Gradient_minimizer minimizer (adapter, adapter, 1, 1, 1, 2);

        minimizer.stop_logic_ = Minimizer::AND;   
        minimizer.line_search_type_ = Minimizer::BRENT; 
        minimizer.set_brack_K (1.618034);        //GOLD //fix
        minimizer.set_gradient_tolerance   (.1);
        minimizer.set_scale                (1.);
        minimizer.set_step_tolerance       (0.01);
        minimizer.set_functional_tolerance (0.001);

        Minimizer::Type type 
            = Minimizer::Hybrid_Lui_Storey_and_Conjugate_Descent;
            //= Minimizer::Steepest_Descent;

        to_user().status ("20 iterations of optimization started");
        //set_run_state ();
        //Lock <Interaction> lock (model.kit().interaction ());
        
        minimizer.conjugate_gradient (type, 20);

        model.kit().visual().renew ();
        //set_stop_state ();

        //model().kit().interaction ().report (&time, "F speed");

        double E = minimizer.func_val ();

        char stop_reason[100];
        minimizer.get_stop_reason (stop_reason);

        Text text = Text() 
            + " E = " + E                           + "     " 
    //                + " grad = " + rms_gradient + "     " 
            + " Grad = " + minimizer.relative_gradient() + "     " 
            + " Iter = " + minimizer.get_iteration()     + "     "
            + " " + stop_reason                     + "     "
            + timer.elapsed() + " sec   ";
        to_user(). status (text);
     
        //energy.flush_cache ();//fix remove

        text = model.kit().file_name ();
        text += "\n";
        text += "Energy ";      text += Text(E, 15);
        text +=  "  Gradient "; text += Text(minimizer.relative_gradient   (), 3);
        text += " Step ";       text += Text(minimizer.step_tolerance      (), 3);
        text += " Function ";   text += Text(minimizer.functional_tolerance(), 3);
        text += " Iteration ";  text += Text(minimizer.get_iteration       (), 3);
        text += "\n";
        text += stop_reason;     text += "   "; 
        text += timer.elapsed(); text += " sec\n";
        
        text += "Call: ";
        text += minimizer.get_function_calls ();         text += " function, "   ;
        text += minimizer.get_gradient_calls ();         text += " gradient, "   ;
        text += minimizer.get_bracketing_func_calls ();  text += " bracketing, " ;
        text += minimizer.get_brent_func_calls ();       text += " Brent, "      ;
        text += minimizer.get_quadratic_func_calls ();   text += " quadratic, "  ;
        text += minimizer.get_cubic_func_calls ();       text += " cubic, "      ;
        text += minimizer.get_other_func_calls ();       text += " others. "     ;

        log() << text << "\n";

    }
    catch(Text & message)           {to_user().error (message);}
    catch(Flaw & exception)         {to_user().error (exception.message());}
    //catch(Contract_violation & exc) {to_user().error (exc.message());}
    catch(...)                      {to_user().fatal_error ("Unknown error.");}
}

void Module_min::
optimize ()
{
    Panel_optimize::singleton()->popup();
    return;

//    Model       & model       = Project::singleton().current_model();
//    Force_field & force_field = model.kit().force_field();
//
//    Lock <Interaction> lock (model.kit().interaction ());
//
//    int size = 3 * model.atom_count(), i;
//    Array <double> gradient;
//    for (i=0;  i<size;  i++)
//        gradient.push_back (0);
//
//    //if (false)
//    {
//    Model_adapter adapter (model);
//    Gradient_minimizer minimize (adapter, adapter, 1, 1, 1, 0.2);
//    minimize.stop_logic_       = Minimizer::AND;        
//    minimize.line_search_type_ = Minimizer::CUBIC;   
//    //minimize.line_search_type_ = Minimizer::BRENT;   
//    minimize.set_brack_K        (1.618034);        //GOLD
//    Timer timer;
//    //minimize.Polak_Ribiere (1000);
//
//    minimize.conjugate_gradient (Minimizer::
//        //Hestenes_Stiefel,     1000);                  // 620  592      72       347            
//        //Lui_Storey,           1000);                  // 638  609   36          568            
//        //Polak_Ribiere,        1000);                  // 617  610               742-b          
//        //Fletcher_Reeves,      1000);                  //1299 1306      70      1408-c          
//        //Dai_Yuan,             1000);                  //1274 1276      71      1428-c          
//        //Conjugate_Descent,    1000);                  //1419 1314   63 55      1325-c          
//                                                                            
//        //Hu_Storey,            1000);                  // 593  610               342            
//        //Hybrid_Dai_Yuan_zero, 1000);                  // 627  594   46  64      778-           
//        Hybrid_Lui_Storey_and_Conjugate_Descent,100);// 595  593   24  52      392            
//                                                        //BRENT CUBIC                   
//                                                        //IL gas      C60 C60-bad Water8-5.5833A 
//
//    model.kit().visual().renew ();
//
//    double E = minimize.func_val ();
//
//    char stop_reason[100];
//    minimize.get_stop_reason (stop_reason);
//
//    Text text = Text() 
//        + " E = " + E                           + "     " 
////                + " grad = " + rms_gradient + "     " 
//        + " Grad = " + minimize.relative_gradient() + "     " 
//        + " Iter = " + minimize.get_iteration()     + "     "
//        + " " + stop_reason                     + "     "
//        + timer.elapsed() + " sec   ";
//    to_user(). status (text);
//
//    //energy.flush_cache ();//fix remove
//
//    log () << "\n" << model.kit().file_name ()                  << "\n";
//    log () << "____________________________________"            << "\n";
//    log ()                                                      << "\n";
//    log () << "Energy           = " << Text(E, 15)              << "\n";
//    log () << "Gradient         = " << Text(minimize.relative_gradient(), 15)
//                                                                << "\n";
//    log () << "Iteration        = " << Text(minimize.get_iteration(), 15)              
//                                                                << "\n";
//    log () << stop_reason                                       << "\n";
//    log () << timer.elapsed() << " sec"                         << "\n";
//    log () << "____________________________________"            << "\n";
//    current_E.write (log());
//    log () << "\n";
//    }
//
//    return;
//
//    {
//    Model_energy energy (model, model, force_field);
//
//    Model_energy_as_differentiable_and_retrievable_coordinates  
//        model_energy (energy);
//
//    Gradient_minimizer mini (model_energy, model_energy, 1, 1, 1, 0.2);
//    mini.stop_logic_       = Minimizer::AND;        
//    mini.line_search_type_ = Minimizer::CUBIC;   
//    //mini.line_search_type_ = Minimizer::BRENT;   
//    mini.set_brack_K        (1.618034);        //GOLD
//
//    Timer timer;
//    //mini.Polak_Ribiere (10);
//    //mini.Ravine (100);
//    mini.conjugate_gradient (Minimizer::Fletcher_Reeves, 1000);
//    
//    /*double E = mini.f_grad (gradient);
//
//    double rms_gradient = 0;
//    for (i=0;  i<size;  i++)
//        rms_gradient += gradient[i] * gradient[i];
//    rms_gradient = sqrt (rms_gradient);//*/
//
//    double E = mini.func_val ();
//
//    char stop_reason[100];
//    mini.get_stop_reason (stop_reason);
//
//    Text text = Text() 
//        + " E = " + E                           + "     " 
////                + " grad = " + rms_gradient + "     " 
//        + " Grad = " + mini.relative_gradient() + "     " 
//        + " Iter = " + mini.get_iteration()     + "     "
//        + " " + stop_reason                     + "     "
//        + timer.elapsed() + " sec   ";
//    to_user(). status (text);
//
//    energy.flush_cache ();//fix remove
//
//    log () << "\n" << model.kit().file_name ()                  << "\n";
//    log () << "____________________________________"            << "\n";
//    log ()                                                      << "\n";
//    log () << "Energy           = " << Text(E, 15)              << "\n";
//    log () << "Gradient         = " << Text(mini.relative_gradient(), 15)
//                                                                << "\n";
//    log () << "Iteration        = " << Text(mini.get_iteration(), 15)              
//                                                                << "\n";
//    log () << stop_reason                                       << "\n";
//    log () << timer.elapsed() << " sec"                         << "\n";
//    log () << "____________________________________"            << "\n";
//    current_E.write (log());
//    log () << "\n";
//    }
}

}//MM








